@@ -1,4 +1,3 @@
-#!/usr/bin/env python
 # Copyright 2008 Orbitz WorldWide
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -25,7 +24,7 @@
 #		Archive = Point+
 #			Point = timestamp,value
 
-import os, struct, time
+import os, struct, time, operator, itertools
 
 try:
   import fcntl
@@ -33,6 +32,42 @@ try:
 except ImportError:
   CAN_LOCK = False
 
+try:
+  import ctypes
+  import ctypes.util
+  CAN_FALLOCATE = True
+except ImportError:
+  CAN_FALLOCATE = False
+
+fallocate = None
+
+if CAN_FALLOCATE: 
+  libc_name = ctypes.util.find_library('c')
+  libc = ctypes.CDLL(libc_name)
+  c_off64_t = ctypes.c_int64
+  c_off_t = ctypes.c_int
+
+  try:
+    _fallocate = libc.posix_fallocate64
+    _fallocate.restype = ctypes.c_int
+    _fallocate.argtypes = [ctypes.c_int, c_off64_t, c_off64_t]
+  except AttributeError, e:
+    try:
+      _fallocate = libc.posix_fallocate
+      _fallocate.restype = ctypes.c_int
+      _fallocate.argtypes = [ctypes.c_int, c_off_t, c_off_t]
+    except AttributeError, e:
+      CAN_FALLOCATE = False
+
+  if CAN_FALLOCATE:
+    def _py_fallocate(fd, offset, len_):
+      res = _fallocate(fd.fileno(), offset, len_)
+      if res != 0:
+        raise IOError(res, 'fallocate')
+    fallocate = _py_fallocate
+  del libc
+  del libc_name
+
 LOCK = False
 CACHE_HEADERS = False
 AUTOFLUSH = False
@@ -64,44 +99,51 @@ aggregationMethods = aggregationTypeToMethod.values()
 debug = startBlock = endBlock = lambda *a,**k: None
 
 UnitMultipliers = {
-  's' : 1,
-  'm' : 60,
-  'h' : 60 * 60,
-  'd' : 60 * 60 * 24,
-  'y' : 60 * 60 * 24 * 365,
+  'seconds' : 1,
+  'minutes' : 60,
+  'hours' : 3600,
+  'days' : 86400,
+  'weeks' : 86400 * 7,
+  'years' : 86400 * 365
 }
 
 
+def getUnitString(s):
+  if 'seconds'.startswith(s): return 'seconds'
+  if 'minutes'.startswith(s): return 'minutes'
+  if 'hours'.startswith(s): return 'hours'
+  if 'days'.startswith(s): return 'days'
+  if 'weeks'.startswith(s): return 'weeks'
+  if 'years'.startswith(s): return 'years'
+  raise ValueError("Invalid unit '%s'" % s)
+
 def parseRetentionDef(retentionDef):
+  import re
   (precision, points) = retentionDef.strip().split(':')
 
   if precision.isdigit():
-    precisionUnit = 's'
-    precision = int(precision)
+    precision = int(precision) * UnitMultipliers[getUnitString('s')]
   else:
-    precisionUnit = precision[-1]
-    precision = int( precision[:-1] )
+    precision_re = re.compile(r'^(\d+)([a-z]+)$')
+    match = precision_re.match(precision)
+    if match:
+      precision = int(match.group(1)) * UnitMultipliers[getUnitString(match.group(2))]
+    else:
+      raise ValueError("Invalid precision specification '%s'" % precision)
 
   if points.isdigit():
-    pointsUnit = None
     points = int(points)
   else:
-    pointsUnit = points[-1]
-    points = int( points[:-1] )
-
-  if precisionUnit not in UnitMultipliers:
-    raise ValueError("Invalid unit: '%s'" % precisionUnit)
-
-  if pointsUnit not in UnitMultipliers and pointsUnit is not None:
-    raise ValueError("Invalid unit: '%s'" % pointsUnit)
-
-  precision = precision * UnitMultipliers[precisionUnit]
-
-  if pointsUnit:
-    points = points * UnitMultipliers[pointsUnit] / precision
+    points_re = re.compile(r'^(\d+)([a-z]+)$')
+    match = points_re.match(points)
+    if match:
+      points = int(match.group(1)) * UnitMultipliers[getUnitString(match.group(2))] / precision
+    else:
+      raise ValueError("Invalid retention specification '%s'" % points)
 
   return (precision, points)
 
+
 class WhisperException(Exception):
     """Base class for whisper exceptions."""
 
@@ -184,7 +226,7 @@ def __readHeader(fh):
     try:
       (offset,secondsPerPoint,points) = struct.unpack(archiveInfoFormat,packedArchiveInfo)
     except:
-      raise CorruptWhisperFile("Unable to read archive %d metadata" % i, fh.name)
+      raise CorruptWhisperFile("Unable to read archive%d metadata" % i, fh.name)
 
     archiveInfo = {
       'offset' : offset,
@@ -249,47 +291,55 @@ aggregationMethod specifies the method to use when propogating data (see ``whisp
 def validateArchiveList(archiveList):
   """ Validates an archiveList.
   An ArchiveList must:
-  1. Have at least one archive config. Example: (60, 86400) 
-  2. No archive may be a duplicate of another. 
+  1. Have at least one archive config. Example: (60, 86400)
+  2. No archive may be a duplicate of another.
   3. Higher precision archives' precision must evenly divide all lower precision archives' precision.
   4. Lower precision archives must cover larger time intervals than higher precision archives.
+  5. Each archive must have at least enough points to consolidate to the next archive
 
   Returns True or False
   """
 
-  try:
-    if not archiveList:
-      raise InvalidConfiguration("You must specify at least one archive configuration!")
+  if not archiveList:
+    raise InvalidConfiguration("You must specify at least one archive configuration!")
 
-    archiveList.sort(key=lambda a: a[0]) #sort by precision (secondsPerPoint)
+  archiveList.sort(key=lambda a: a[0]) #sort by precision (secondsPerPoint)
 
-    for i,archive in enumerate(archiveList):
-      if i == len(archiveList) - 1:
-        break
+  for i,archive in enumerate(archiveList):
+    if i == len(archiveList) - 1:
+      break
 
-      next = archiveList[i+1]
-      if not (archive[0] < next[0]):
-        raise InvalidConfiguration("You cannot configure two archives "
-          "with the same precision %s,%s" % (archive,next))
+    nextArchive = archiveList[i+1]
+    if not archive[0] < nextArchive[0]:
+      raise InvalidConfiguration("A Whisper database may not configured having"
+        "two archives with the same precision (archive%d: %s, archive%d: %s)" %
+        (i, archive, i + 1, nextArchive))
 
-      if (next[0] % archive[0]) != 0:
-        raise InvalidConfiguration("Higher precision archives' precision "
-          "must evenly divide all lower precision archives' precision %s,%s" \
-          % (archive[0],next[0]))
+    if nextArchive[0] % archive[0] != 0:
+      raise InvalidConfiguration("Higher precision archives' precision "
+        "must evenly divide all lower precision archives' precision "
+        "(archive%d: %s, archive%d: %s)" %
+        (i, archive[0], i + 1, nextArchive[0]))
 
-      retention = archive[0] * archive[1]
-      nextRetention = next[0] * next[1]
+    retention = archive[0] * archive[1]
+    nextRetention = nextArchive[0] * nextArchive[1]
 
-      if not (nextRetention > retention):
-        raise InvalidConfiguration("Lower precision archives must cover "
-          "larger time intervals than higher precision archives %s,%s" \
-          % (archive,next))
-  
-  except:
-    return False
-  return True
+    if not nextRetention > retention:
+      raise InvalidConfiguration("Lower precision archives must cover "
+        "larger time intervals than higher precision archives "
+        "(archive%d: %s seconds, archive%d: %s seconds)" %
+        (i, retention, i + 1, nextRetention))
 
-def create(path,archiveList,xFilesFactor=None,aggregationMethod=None):
+    archivePoints = archive[1]
+    pointsPerConsolidation = nextArchive[0] / archive[0]
+    if not archivePoints >= pointsPerConsolidation:
+      raise InvalidConfiguration("Each archive must have at least enough points "
+        "to consolidate to the next archive (archive%d consolidates %d of "
+        "archive%d's points but it has only %d total points)" %
+        (i + 1, pointsPerConsolidation, i, archivePoints))
+
+
+def create(path,archiveList,xFilesFactor=None,aggregationMethod=None,sparse=False,useFallocate=False):
   """create(path,archiveList,xFilesFactor=0.5,aggregationMethod='average')
 
 path is a string
@@ -304,9 +354,7 @@ aggregationMethod specifies the function to use when propogating data (see ``whi
     aggregationMethod = 'average'
 
   #Validate archive configurations...
-  validArchive = validateArchiveList(archiveList)
-  if not validArchive:
-    raise InvalidConfiguration("There was a problem creating %s due to an invalid schema config." % path) 
+  validateArchiveList(archiveList)
 
   #Looks good, now we create the file and write the header
   if os.path.exists(path):
@@ -317,7 +365,7 @@ aggregationMethod specifies the function to use when propogating data (see ``whi
     fcntl.flock( fh.fileno(), fcntl.LOCK_EX )
 
   aggregationType = struct.pack( longFormat, aggregationMethodToType.get(aggregationMethod, 1) )
-  oldest = sorted([secondsPerPoint * points for secondsPerPoint,points in archiveList])[-1]
+  oldest = max([secondsPerPoint * points for secondsPerPoint,points in archiveList])
   maxRetention = struct.pack( longFormat, oldest )
   xFilesFactor = struct.pack( floatFormat, float(xFilesFactor) )
   archiveCount = struct.pack(longFormat, len(archiveList))
@@ -331,8 +379,22 @@ aggregationMethod specifies the function to use when propogating data (see ``whi
     fh.write(archiveInfo)
     archiveOffsetPointer += (points * pointSize)
 
-  zeroes = '\x00' * (archiveOffsetPointer - headerSize)
-  fh.write(zeroes)
+  #If configured to use fallocate and capable of fallocate use that, else
+  #attempt sparse if configure or zero pre-allocate if sparse isn't configured.
+  if CAN_FALLOCATE and useFallocate:
+    remaining = archiveOffsetPointer - headerSize
+    fallocate(fh, headerSize, remaining)
+  elif sparse:
+    fh.seek(archiveOffsetPointer - 1)
+    fh.write('\x00')
+  else:
+    remaining = archiveOffsetPointer - headerSize
+    chunksize = 16384
+    zeroes = '\x00' * chunksize
+    while remaining > chunksize:
+      fh.write(zeroes)
+      remaining -= chunksize
+    fh.write(zeroes[:remaining])
 
   if AUTOFLUSH:
     fh.flush()
@@ -340,7 +402,7 @@ aggregationMethod specifies the function to use when propogating data (see ``whi
 
   fh.close()
 
-def __aggregate(aggregationMethod, knownValues):
+def aggregate(aggregationMethod, knownValues):
   if aggregationMethod == 'average':
     return float(sum(knownValues)) / float(len(knownValues))
   elif aggregationMethod == 'sum':
@@ -414,7 +476,7 @@ def __propagate(fh,header,timestamp,higher,lower):
 
   knownPercent = float(len(knownValues)) / float(len(neighborValues))
   if knownPercent >= xff: #we have enough data to propagate a value!
-    aggregateValue = __aggregate(aggregationMethod, knownValues)
+    aggregateValue = aggregate(aggregationMethod, knownValues)
     myPackedPoint = struct.pack(pointFormat,lowerIntervalStart,aggregateValue)
     fh.seek(lower['offset'])
     packedPoint = fh.read(pointSize)
@@ -559,6 +621,7 @@ def __archive_update_many(fh,header,archive,points):
   step = archive['secondsPerPoint']
   alignedPoints = [ (timestamp - (timestamp % step), value)
                     for (timestamp,value) in points ]
+  alignedPoints = dict(alignedPoints).items() # Take the last val of duplicates
   #Create a packed string for each contiguous sequence of points
   packedStrings = []
   previousInterval = None
@@ -637,7 +700,12 @@ def fetch(path,fromTime,untilTime=None):
 
 path is a string
 fromTime is an epoch time
-untilTime is also an epoch time, but defaults to now
+untilTime is also an epoch time, but defaults to now.
+
+Returns a tuple of (timeInfo, valueList)
+where timeInfo is itself a tuple of (fromTime, untilTime, step)
+
+Returns None if no data can be returned
 """
   fh = open(path,'rb')
   return file_fetch(fh, fromTime, untilTime)
@@ -651,16 +719,25 @@ def file_fetch(fh, fromTime, untilTime):
   fromTime = int(fromTime)
   untilTime = int(untilTime)
 
+  # Here we try and be flexible and return as much data as we can.
+  # If the range of data is from too far in the past or fully in the future, we
+  # return nothing
+  if (fromTime > untilTime):
+    raise InvalidTimeInterval("Invalid time interval: from time '%s' is after until time '%s'" % (fromTime, untilTime))
+
   oldestTime = now - header['maxRetention']
+  # Range is in the future
+  if fromTime > now:
+    return None
+  # Range is beyond retention
+  if untilTime < oldestTime:
+    return None
+  # Range requested is partially beyond retention, adjust
   if fromTime < oldestTime:
     fromTime = oldestTime
-
-  if not (fromTime < untilTime):
-    raise InvalidTimeInterval("Invalid time interval")
+  # Range is partially in the future, adjust
   if untilTime > now:
     untilTime = now
-  if untilTime < fromTime:
-    untilTime = now
 
   diff = now - fromTime
   for archive in header['archives']:
@@ -723,3 +800,30 @@ def file_fetch(fh, fromTime, untilTime):
   fh.close()
   timeInfo = (fromInterval,untilInterval,step)
   return (timeInfo,valueList)
+
+def merge(path_from, path_to, step=1<<12):
+  headerFrom = info(path_from)
+
+  archives = headerFrom['archives']
+  archives.sort(key=operator.itemgetter('retention'), reverse=True)
+
+  # Start from maxRetention of the oldest file, and skip forward at max 'step'
+  # points at a time.
+  fromTime = int(time.time()) - headerFrom['maxRetention']
+  for archive in archives:
+    pointsRemaining = archive['points']
+    while pointsRemaining:
+      pointsToRead = step
+      if pointsRemaining < step:
+        pointsToRead = pointsRemaining
+      pointsRemaining -= pointsToRead
+      untilTime = fromTime + (pointsToRead * archive['secondsPerPoint'])
+      (timeInfo, values) = fetch(path_from, fromTime, untilTime)
+      (start, end, archive_step) = timeInfo
+      pointsToWrite = list(itertools.ifilter(
+        lambda points: points[1] is not None,
+        itertools.izip(xrange(start, end, archive_step), values)))
+      pointsToWrite.sort(key=lambda p: p[0],reverse=True) #order points by timestamp, newest first
+      update_many(path_to, pointsToWrite)
+      fromTime = untilTime
+#!/usr/bin/env python
